"""Package init."""
from . import distributions
from .version import __version__
from .losses import elbo, max_posterior
from .baselayers import stack
from .layers import (Activation, DropOut, MaxPool2D, Reshape, DenseVariational,
                     DenseMAP, InputLayer, EmbedVariational, RandomFourier,
                     RandomArcCosine)
from .hlayers import Concat, Sum, PerFeature
from .impute import (MeanImpute, FixedNormalImpute, LearnedScalarImpute,
                     LearnedNormalImpute)
from .kernels import RBF, Matern, RBFVariational
from .distributions import (norm_prior, norm_posterior, gaus_posterior)
from .util import (batch, pos, predict_expected, predict_samples,
                   batch_prediction)
from .random import set_hyperseed

__all__ = (
    'distributions',
    '__version__',
    'elbo',
    'max_posterior',
    'Activation',
    'DropOut',
    'MaxPool2D',
    'Reshape',
    'DenseVariational',
    'DenseMAP',
    'EmbedVariational',
    'RandomFourier',
    'RandomArcCosine',
    'norm_prior',
    'norm_posterior',
    'gaus_posterior',
    'batch',
    'pos',
    'predict_expected',
    'predict_samples',
    'batch_prediction',
    'set_hyperseed',
    'InputLayer',
    'stack',
    'Sum',
    'Concat',
    'PerFeature',
    'MeanImpute',
    'FixedNormalImpute',
    'LearnedScalarImpute',
    'LearnedNormalImpute',
    'RBF',
    'RBFVariational',
    'Matern'
)



"""Test the aboleth utilities."""

from types import GeneratorType

import numpy as np
import tensorflow as tf

import aboleth as ab


def test_batch():
    """Test the batch feed dict generator."""
    X = np.arange(100)
    fd = {'X': X}

    data = ab.batch(fd, batch_size=10, n_iter=10)

    # Make sure this is a generator
    assert isinstance(data, GeneratorType)

    # Make sure we get a dict back of a length we expect
    d = next(data)
    assert isinstance(d, dict)
    assert 'X' in d
    assert len(d['X']) == 10

    # Test we get all of X back in one sweep of the data
    accum = list(d['X'])
    for ds in data:
        assert len(ds['X']) == 10
        accum.extend(list(ds['X']))

    assert len(accum) == len(X)
    assert set(X) == set(accum)


def test_batch_predict():
    """Test the batch prediction feed dict generator."""
    X = np.arange(100)
    fd = {'X': X}

    data = ab.batch_prediction(fd, batch_size=10)

    # Make sure this is a generator
    assert isinstance(data, GeneratorType)

    # Make sure we get a dict back of a length we expect with correct indices
    for ind, d in data:
        assert isinstance(d, dict)
        assert 'X' in d
        assert len(d['X']) == 10
        assert all(X[ind] == d['X'])


def test_predict_samples():
    """Test the predict_samples aggregator."""
    X = np.ones((10, 100, 1), dtype=np.float32)
    X_ = tf.placeholder(tf.float32, (10, None, 1))
    Xt = tf.identity(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        samps = ab.predict_samples(Xt, {X_: X}, n_groups=10)  # 10 replicates
        assert samps.shape == (100, 100, 1)
        assert np.allclose(samps, 1.)


def test_predict_expected():
    """Test the predict_expected computation."""
    X = np.ones((10, 100, 1), dtype=np.float32)
    X_ = tf.placeholder(tf.float32, (10, None, 1))
    Xt = tf.identity(X_)

    tc = tf.test.TestCase()
    with tc.test_session():
        samps = ab.predict_expected(Xt, {X_: X}, n_groups=10)  # 10 replicates
        assert samps.shape == (100, 1)
        assert np.allclose(samps, np.ones((100, 1)))  # test average on axis 0



"""Tests for imputation layers."""

import numpy as np
import tensorflow as tf

import aboleth as ab


def test_mean_impute(make_missing_data):
    """Test the impute_mean."""
    _, m, X, _ = make_missing_data

    # This replicates the input layer behaviour
    def data_layer(**kwargs):
        return kwargs['X'], 0.0

    def mask_layer(**kwargs):
        return kwargs['M'], 0.0

    impute = ab.MeanImpute(data_layer, mask_layer)

    F, KL = impute(X=X, M=m)

    tc = tf.test.TestCase()
    with tc.test_session():
        X_imputed = F.eval()
        imputed_data = X_imputed[1, m]
        assert np.allclose(imputed_data[-5:], [1., 2., 3., 4., 5.])
        assert KL.eval() == 0.0


def test_fixed_gaussian_impute(make_missing_data):
    """Test the impute_mean."""
    ab.set_hyperseed(100)
    _, m, X, _ = make_missing_data

    # This replicates the input layer behaviour
    def data_layer(**kwargs):
        return kwargs['X'], 0.0

    def mask_layer(**kwargs):
        return kwargs['M'], 0.0

    n, N, D = X.shape
    mean_array = 2 * np.ones(D).astype(np.float32)
    std_array = np.sqrt(0.001) * np.ones(D).astype(np.float32)
    impute = ab.FixedNormalImpute(data_layer, mask_layer, mean_array,
                                  std_array)

    F, KL = impute(X=X, M=m)

    tc = tf.test.TestCase()
    with tc.test_session():
        X_imputed = F.eval()
        imputed_data = X_imputed[1, m]
        correct = np.array([1.94, 1.97,  1.93,  2.03, 2.02])
        assert np.allclose(imputed_data[-5:], correct, atol=0.1)
        assert KL.eval() == 0.0


def test_learned_scalar_impute(make_missing_data):
    """Test the impute that learns a scalar value to impute for each col."""
    ab.set_hyperseed(100)
    _, m, X, _ = make_missing_data

    # This replicates the input layer behaviour
    def data_layer(**kwargs):
        return kwargs['X'], 0.0

    def mask_layer(**kwargs):
        return kwargs['M'], 0.0

    n, N, D = X.shape
    impute = ab.LearnedScalarImpute(data_layer, mask_layer)

    F, KL = impute(X=X, M=m)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        X_imputed = F.eval()
        assert KL.eval() == 0.0  # Might want to change this in the future
        assert(X_imputed.shape == X.shape)


def test_learned_normal_impute(make_missing_data):
    """Test the learned normal impute function."""
    ab.set_hyperseed(100)
    _, m, X, _ = make_missing_data

    # This replicates the input layer behaviour
    def data_layer(**kwargs):
        return kwargs['X'], 0.0

    def mask_layer(**kwargs):
        return kwargs['M'], 0.0

    n, N, D = X.shape
    impute = ab.LearnedNormalImpute(data_layer, mask_layer)

    F, KL = impute(X=X, M=m)

    tc = tf.test.TestCase()
    with tc.test_session():
        tf.global_variables_initializer().run()
        X_imputed = F.eval()
        assert KL.eval() == 0.0  # Might want to change this in the future
        assert(X_imputed.shape == X.shape)
